/* 
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <time.h>
#include <locale.h>
#include "spoof_qt.h"
#include <QCommandLineParser>
#include <QtGlobal>
#include <QDir>
#include "../../config.h"
#include "common.h"
static const char cvsid[] USED = "$Id: common.cpp,v 1.35 2016/05/18 18:09:39 kkeys Exp $";

SpooferBase::OnDemandDevice SpooferBase::outdev(stdout);
SpooferBase::OnDemandDevice SpooferBase::errdev(stderr);
QTextStream SpooferBase::spout(&SpooferBase::outdev);
QTextStream SpooferBase::sperr(&SpooferBase::errdev);
bool SpooferBase::optDebug = false;
QString SpooferBase::optSettings;
SpooferBase::Config *SpooferBase::config;
QSettings *SpooferBase::Config::settings;
QList<SpooferBase::Config::MemberBase*> SpooferBase::Config::members;

// We avoid ".log" Suffix because OSX "open" would launch a log reader app
// we don't want.
const QString SpooferBase::proberLogFtime = QSL("'spoofer-prober-'yyyy~MM~dd-HH~mm~ss'.txt'");
const QString SpooferBase::proberLogGlob = QSL("spoofer-prober-\?\?\?\?\?\?\?\?-\?\?\?\?\?\?.txt"); // backslashes prevent trigraphs
const QString SpooferBase::proberLogRegex = QSL("spoofer-prober-(\\d{4})(\\d{2})(\\d{2})-(\\d{2})(\\d{2})(\\d{2}).txt$");

SpooferBase::SpooferBase()
{
    setlocale(LC_NUMERIC, "C");
    appDir = QCoreApplication::applicationDirPath();
    appFile = QCoreApplication::applicationFilePath();

    // for QStandardPaths::standardLocations() and QSettings
    QCoreApplication::setOrganizationName(QSL(ORG_NAME));
    QCoreApplication::setOrganizationDomain(QSL(ORG_DOMAIN));
    QCoreApplication::setApplicationName(QSL("Spoofer")); // may change later

    qInstallMessageHandler(logHandler);
}

qint64 SpooferBase::OnDemandDevice::writeData(const char *data, qint64 maxSize)
{
    if (newdev) {
	QFile *file = dynamic_cast<QFile*>(dev);
	QFile *newfile = dynamic_cast<QFile*>(newdev);
	if (newfile) {
	    // Make sure filename is clean and absolute for comparison below.
	    QDir newpath(QDir::cleanPath(newfile->fileName()));
	    newfile->setFileName(newpath.absolutePath());
	}
	if (file && file->isOpen() && newfile && file->fileName() ==
	    newfile->fileName())
	{
	    // Old dev is open, and newdev is the same file; ignore newdev.
	} else {
	    char buf[2048];
	    if (dev && dev->isOpen() && !newname.isEmpty()) {
		snprintf(buf, sizeof(buf), "Redirecting output to %s\n",
		    qPrintable(newname));
		dev->write(buf);
	    }
	    if (!newdev->open(WriteOnly|Unbuffered|Append|Text)) {
		if (dev && dev->isOpen()) {
		    snprintf(buf, sizeof(buf), "Redirection failed: %s.\n",
			qPrintable(newdev->errorString()));
		    dev->write(buf);
		}
		delete newdev;
	    } else { // success
		if (dev) delete dev;
		dev = newdev;
	    }
	}
	newdev = nullptr;
    }

    if (!dev) {
	if (!fallback) {
	    setErrorString(QSL("output device is not set"));
	    return maxSize; // *dev failed, but *this can still work
	}
	// E.g., dev was nulled in a previous call, and fallback was set later.
	newdev = fallback;
	fallback = nullptr;
	return writeData(data, maxSize); // Try again with the fallback.
    }

    qint64 retval = dev->write(data, maxSize);
    if (retval < 0) {
	// E.g., stderr is closed when running as a Windows service.
	if (!fallback) { // There was no fallback.
	    delete dev; dev = nullptr;
	    return maxSize; // *dev failed, but *this can still work
	}
	newdev = fallback;
	fallback = nullptr;
	return writeData(data, maxSize); // Try again with the fallback.
    }
    fallback = nullptr; // Dev worked; we won't need the fallback.
    return retval;
}

void SpooferBase::logHandler(QtMsgType type, const QMessageLogContext &ctx,
    const QString &msg)
{
    Q_UNUSED(ctx);

    if (!(DEBUG-0) && type == QtDebugMsg)
	return;

    const char *prefix =
	(type == QtDebugMsg)    ? "Debug"    :
	(type == QtWarningMsg)  ? "Warning"  :
	(type == QtCriticalMsg) ? "Critical" :
	(type == QtFatalMsg)    ? "Fatal"    :
	"Info";

    static int depth = 0;
    if (++depth > 1) { // should not happen
	fprintf(stderr, "INTERNAL ERROR: logHandler recursion\n");
	fprintf(stderr, "%s: %s\n", prefix, qPrintable(msg));
	::fflush(stderr);
    } else {
	sperr << prefix << ": " << msg << endl;
    }
    depth--;
}

// Caller can add additional options before calling parseCommandLine(), and
// inspect them after.
bool SpooferBase::parseCommandLine(QCommandLineParser &clp, QString desc)
{
    clp.setApplicationDescription(desc);

    QCommandLineOption cloDebug(QSL("debug"), QSL("Use alternate debugging copy of settings."));
    clp.addOption(cloDebug);

    QCommandLineOption cloSettings(QSL("settings"), QSL("Read settings from <file>."),
	QSL("file"));
    clp.addOption(cloSettings);

    QCommandLineOption cloHelp = clp.addHelpOption();

    if (!clp.parse(QCoreApplication::arguments())) {
	sperr << clp.errorText().toStdString().c_str() << endl;
	return false;
    }

    if (clp.isSet(cloHelp)) {
	sperr << clp.helpText().toStdString().c_str() << endl;
	return false;
    }

    optDebug = clp.isSet(cloDebug);
    if (optDebug)
	QCoreApplication::setApplicationName(QSL("Spoofer-debug"));

    if (clp.isSet(cloSettings))
	optSettings = clp.value(cloSettings);

    return true;
}

// Format is like QDateTime::toString(), with the addition that '~' will be
// removed after formatting, allowing you to create adjacent non-separated
// fields in output by inserting '~' between them in the input format.
QString SpooferBase::ftime(QString fmt, const time_t *tp)
{
    QDateTime qdt(tp ? QDateTime::fromTime_t(*tp, Qt::LocalTime) :
	QDateTime::currentDateTime());
    return qdt.toString(!fmt.isEmpty() ? fmt : QSL("yyyy-MM-dd HH:mm:ss t")).remove(QLatin1Char('~'));
}

QString SpooferBase::ftime_utc(QString fmt, const time_t *tp)
{
    QDateTime qdt(tp ? QDateTime::fromTime_t(*tp, Qt::UTC) :
	QDateTime::currentDateTimeUtc());
    return qdt.toString(!fmt.isEmpty() ? fmt : QSL("yyyy-MM-dd HH:mm:ss t")).remove(QLatin1Char('~'));
}

