/* 
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string>
#include <sstream>
#include <exception>
#include <stdexcept>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include "spoof_qt.h"
#include <QDebug>
#include <QtDebug>
#include <QCommandLineParser>
#include <QLocalServer>
#include <QLocalSocket>
#include <QProcess>
#include <QStandardPaths>
#include <QDir>
#include <QNetworkConfigurationManager> // experimental
#include <QNetworkInterface> // experimental
#include <QCoreApplication>
#include <QLockFile>
#include <QRegularExpression>
#include "../../config.h"
#include "app.h"

#ifdef Q_OS_UNIX
#include <signal.h>
#endif

static const char cvsid[] USED = "$Id: app.cpp,v 1.138 2016/05/26 18:49:04 kkeys Exp $";

const QString App::name(QSL(APPNAME));
int App::argc = 0;
char **App::argv = nullptr;
QString App::dataDir;

#if defined(Q_OS_WIN32)
#   include "appwin.h"
    App *App::newapp(int &argc, char **argv) { return new AppWin(argc, argv); }
#elif defined(Q_OS_MAC) // must come before Q_OS_UNIX
#   include "appmac.h"
    App *App::newapp(int &argc, char **argv) { return new AppMac(argc, argv); }
#elif defined(Q_OS_UNIX)
#   include "appunix.h"
    App *App::newapp(int &argc, char **argv) { return new AppUnix(argc, argv); }
#endif

QString App::AppLog::makeName()
{
    QDir dir(dataDir);
    return dir.filePath(App::name % ftime_utc(QSL("-yyyy~MM~dd-HH~mm~ss'.txt'")));
}

QList<SubnetAddr> App::getAddresses() const
{
    QList<SubnetAddr> addrs;
    QList<QNetworkInterface> ifaces = QNetworkInterface::allInterfaces();
    foreach (QNetworkInterface iface, ifaces) {
	if (!iface.isValid()) continue;
	if (!(iface.flags() & QNetworkInterface::IsRunning)) continue;
	if (!(iface.flags() & QNetworkInterface::IsUp)) continue;
	if ((iface.flags() & QNetworkInterface::IsLoopBack)) continue;
	QList<QNetworkAddressEntry> addrentries = iface.addressEntries();
	foreach (QNetworkAddressEntry addrentry, addrentries) {
	    QHostAddress addr = addrentry.ip();
	    if (addr.isNull()) continue;
	    if (addr.isLoopback()) continue;
	    if (!addr.scopeId().isEmpty()) continue; // link-local or site-local
	    if (addr.protocol() == QAbstractSocket::IPv4Protocol) {
		if (!config->enableIPv4()) continue;
		if (addrentry.prefixLength() == 32)
		    continue; // skip address with full-length mask
	    } else if (addr.protocol() == QAbstractSocket::IPv6Protocol) {
		if (!config->enableIPv6()) continue;
		if (addrentry.prefixLength() == 128)
		    continue; // skip address with full-length mask
	    }
	    addrs.push_back(SubnetAddr(addr, addrentry.prefixLength()));
	}
    }
    return addrs;
}

SubnetAddr SubnetAddr::prefix() const
{
    if (addr().protocol() == QAbstractSocket::IPv4Protocol) {
	qint32 raw = addr().toIPv4Address();
	raw &= (0xFFFFFFFF << (32 - pfxlen()));
	return SubnetAddr(QHostAddress(raw), pfxlen());

    } else if (addr().protocol() == QAbstractSocket::IPv6Protocol) {
	Q_IPV6ADDR raw = addr().toIPv6Address();
	for (int i = 0, n = pfxlen(); i < 16; i++, n -= 8) {
	    if (n <= 0) {
		raw[i] = 0;
	    } else if (n < 8) {
		raw[i] &= 0xFF << (8-n);
	    }
	}
	return SubnetAddr(QHostAddress(raw), pfxlen());
    } else {
	return SubnetAddr(QHostAddress(), 0);
    }
}

void App::handleNetChange()
{
    qDebug() << "possible network change detected" << qPrintable(ftime_utc());
    proberTimer.stop(); // Temporarily prevent prober from starting.
    // Do NOT reset nextProberStart.when here.  If scheduleNextProber() finds
    // no relevant change in the network config, it will restart the timer to
    // resume the original schedule.
    if (prober) return; // prober is running now
    // Wait 1s (in case of a burst of changes), then scheduleNextProber()
    netPollTimer.start(1000*1);
}

void App::scheduleNextProber()
{
    time_t now, start = 0;
    if (prober) return; // prober is running now
    if (paused) return;
    time(&now);
    QList<SubnetAddr> addrs = getAddresses();
    if (nextProberStart.when && addrs == scheduledAddrs) {
	// No change.  Resume original schedule.
	qDebug() << "scheduleNextProber: no relevant change in network config";
	start = nextProberStart.when;
	proberTimer.start(start <= now ? 1 : 1000 * (start - now));
	return;
    }
    proberTimer.stop();
    netPollTimer.stop();
    scheduledAddrs = addrs;
    if (addrs.isEmpty()) {
	// TODO: if net change detection is perfect, don't start netPollTimer
	netPollTimer.start(1000 * config->netPollInterval());
	goto notify;
    }
    qDebug() << "scheduleNextProber" << qPrintable(ftime_utc(QString(), &now));
    foreach (SubnetAddr addr, addrs) {
	const RunRecord &rr = pastRuns[addr.prefix()];
	time_t interval = config->proberInterval();
	if (rr.errors > 0) {
	    // exponential backoff, capped at proberInterval
	    interval = config->proberRetryInterval();
	    for (unsigned i = 1; i < rr.errors; i++) {
		interval <<= 1;
		if (interval >= config->proberInterval()) {
		    interval = config->proberInterval();
		    break;
		}
	    }
	}
	time_t t = rr.t + interval;
	if (!start || t < start) start = t;
    }
    qDebug() << "  earliest" << qPrintable(ftime_utc(QString(), &start));
    if (start < now + config->delayInterval())
	start = now + config->delayInterval();
    proberTimer.start(1000 * (start - now));
    if (start > now + config->delayInterval()) {
	// Re-check the network configuration <delayInterval> before the next
	// scheduled run, or after <netPollInterval>, whichever is earlier.
	// TODO: If net change detection is perfect, don't start netPollTimer.
	netPollTimer.start(1000 * qMin((time_t)config->netPollInterval(),
	    start - now - config->delayInterval()));
    }

notify:
    QString label = start ? ftime_utc(QString(), &start) : QSL("never");
    if (start != nextProberStart.when) {
	sperr << "next prober run: " << label << endl;
	nextProberStart.when = start;
	foreach (QLocalSocket *ui, uiSet) {
	    if (opAllowed(ui, config->unprivView))
		QDataStream(ui) << (qint32)SC_SCHEDULED << nextProberStart;
	}
    } else {
	qDebug() << "next prober run:" << label << "(silent)";
    }
}

void App::dumpNetCfg()
{
    QList<SubnetAddr> addrs = getAddresses();

    foreach (SubnetAddr addr, addrs) {
	qDebug().nospace() << "    " <<
	    (addr.addr().protocol() == QAbstractSocket::IPv4Protocol ? "ipv4" :
	     addr.addr().protocol() == QAbstractSocket::IPv6Protocol ? "ipv6" :
	     "????") << ": " <<
	    qPrintable(addr.toString()) << "  " << qPrintable(addr.prefix().toString());
    }
}

bool App::parseCommandLine(QCommandLineParser &clp)
{
    QCommandLineOption cloDetached(QStringList() << QSL("D") <<
#ifdef Q_OS_WIN32
	QSL("detach"), QSL("detach from console, if there is one")
#else
	QSL("daemon"), QSL("run as a daemon")
#endif
	);
    clp.addOption(cloDetached);

    QCommandLineOption cloShareData(QStringList() << QSL("s") << QSL("share"),
	QSL("allow/don't allow data to be shared publicly"), QSL("1|0"));
    clp.addOption(cloShareData);

    QCommandLineOption cloStartPaused(QStringList() << QSL("p") << QSL("paused"),
	QSL("start with scheduled prober runs disabled"));
    clp.addOption(cloStartPaused);

    QCommandLineOption cloInitOnly(QSL("init"), QSL("initialize only, then exit"));
    clp.addOption(cloInitOnly);

    QCommandLineOption cloDumpPaths(QSL("dump-paths"),
	QSL("dump list of QStandardPaths, for debugging"));
    clp.addOption(cloDumpPaths);

    QCommandLineOption cloStandalone(QSL("standalone"),
	QSL("standalone mode (without a server), for debugging"));
    clp.addOption(cloStandalone);

    QCommandLineOption cloLogfile(QStringList() << QSL("l") << QSL("logfile"),
	QSL("log messages to ") % App::name % QSL("-<time>.txt in the data directory"));
    clp.addOption(cloLogfile);

    if (!SpooferBase::parseCommandLine(clp, QSL("Spoofer scheduler service")))
	return false;

    this->paused = clp.isSet(cloStartPaused);
    this->optDumpPaths = clp.isSet(cloDumpPaths);
    this->optDetach = clp.isSet(cloDetached);
    this->optStandalone = clp.isSet(cloStandalone);
    this->optLogfile = clp.isSet(cloLogfile);
    this->optInitOnly = clp.isSet(cloInitOnly);

    if (clp.isSet(cloShareData)) {
	QString value = clp.value(cloShareData);
	if (value == QSL("0"))
	    optShareData = 0;
	else if (value == QSL("1"))
	    optShareData = 1;
	else {
	    sperr << "illegal option value " << value << endl;
	    return false;
	}
    }

    return true;
}

void App::dumpPaths(void) const
{
    spout << "applicationDirPath: " << "\"" << appDir << "\"" << endl;
    spout << "applicationFilePath: " << "\"" << appFile << "\"" << endl;
    spout << "homePath: " << "\"" << QDir::homePath() << "\"" << endl;

#define dumpQtPaths(id) \
    do { \
	QStandardPaths::StandardLocation type = QStandardPaths:: id ## Location; \
	spout << (int)type << " " << #id << " # " << QStandardPaths::displayName(type) << endl; \
	QStringList list = QStandardPaths::standardLocations(type); \
	for (int i = 0; i < list.size(); ++i) { \
	    spout << "    \"" << list.at(i) << "\"" << endl; \
	} \
    } while (0)

    dumpQtPaths(Temp          );          // 7
    dumpQtPaths(Home          );          // 8
    dumpQtPaths(Data          );          // 9
    dumpQtPaths(GenericData   );          // 11
    dumpQtPaths(Runtime       );          // 12
    dumpQtPaths(Config        );          // 13
    dumpQtPaths(GenericConfig );          // 16
}

bool App::chooseDataDir()
{
    // QStandardPaths::standardLocations(QStandardPaths::DataLocation) may
    // return multiple types of locations, in no particular order:
    //   1 global:    linux example:  /usr/share/CAIDA/Spoofer        
    //                mac example:    /Library/Application Support/CAIDA/Spoofer
    //   2 app base:  mac example:    /Applications/Spoofer.app
    //   3 user:      linux example:  $HOME/.local/share/CAIDA/Spoofer
    // We prefer them in the order listed above (or reversed if debugging).
    {
	int best_score = -1;
	QString appBase = appDir;
#ifdef Q_OS_MAC
	appBase.remove(QSL("/Contents/MacOS"));
#endif
	QStringList paths =
	    QStandardPaths::standardLocations(QStandardPaths::DataLocation);
	qDebug() << "homePath:" << QDir::homePath();
	qDebug() << "rootPath:" << QDir::rootPath();
	foreach (QString path, paths) {
	    if (path.isEmpty()) continue;
	    QString cleanpath = QDir::cleanPath(path);
	    int score = 1;
	    if (cleanpath.startsWith(appBase)) score = 2;
	    else if (QDir::homePath() != QDir::rootPath() &&
		cleanpath.startsWith(QDir::homePath()))
		    score = 3;
#ifdef Q_OS_MAC
	    else if (cleanpath.startsWith(QSL("/private/")))
		score = 3;
#endif
	    qDebug() << "dataDir option, score" << score << ":" << cleanpath;
	    if (dataDir.isEmpty() ||
		(optDebug ? (score > best_score) : (score < best_score)))
	    {
		best_score = score;
		dataDir = cleanpath;
	    }
	}
    }
    if (dataDir.isEmpty()) {
	sperr << "can't determine a suitable data folder" << endl;
	return false;
    }
    return true;
}

QLocalServer *App::listen(bool privileged)
{
    QLocalServer *server = new QLocalServer(this);
    connect(server, &QLocalServer::newConnection, this, &App::uiAccept);
    
    server->setSocketOptions(
#ifndef EVERYONE_IS_PRIVILEGED
	privileged ? QLocalServer::UserAccessOption :
#endif
	QLocalServer::WorldAccessOption);

    QString serverName = QSL("spoofer-") % QString::number(applicationPid());
    if (!privileged) serverName.append(QSL("o"));
    if (!server->listen(serverName)) {
	sperr << "listen: " << server->errorString() << endl;
	delete server;
	return nullptr;
    }
    sperr << "Scheduler listening for " <<
#ifdef EVERYONE_IS_PRIVILEGED
	(privileged ? "privileged " : "unprivileged ") <<
#endif
	"connections on " << server->fullServerName() << endl;
    return server;
}

// Returns true if caller should continue with app.exec().
bool App::init(int &exitCode)
{
    exitCode = 1; // default result is failure
    connect(this, &QCoreApplication::aboutToQuit, this, &App::cleanup);
    QCommandLineParser clp;

    if (!parseCommandLine(clp)) return false;

    // after parsing options, before changing working directory
    if (!initConfig()) return false;

    if (optDumpPaths) dumpPaths();

    if (!optSettings.isEmpty() && !config->dataDir().isEmpty()) {
	// custom settings file with "dataDir" entry
	dataDir = config->dataDir();
    } else {
	// standard settings, or custom settings without "dataDir" entry
	if (!chooseDataDir()) return false;
	qDebug() << "dataDir:" << dataDir;
    }

    QDir dir;
    dir.mkpath(dataDir); // make sure it exists
    QDir::setCurrent(dataDir);
    // don't "config->dataDir(dataDir)" until we have lock and are listening

    // Maybe open new log file using new value of dataDir
    if (optLogfile || errdev.type() == typeid(AppLog)) {
	errdev.setDevice(new AppLog());
    } else {
	errdev.setFallbackDevice(new AppLog());
    }

    if (optShareData >= 0) {
	qDebug() << "config->shareData" << (optShareData == 1);
	config->shareData(optShareData == 1);
    }

    if (optInitOnly) {
	config->dataDir(dataDir);
	config->sync();
	exitCode = 0; // success
	return false; // caller should not do app.exec()
    }

    if (!prestart(exitCode)) return false;

    // Make sure we are the only running instance
    // (after prestart, so we have the correct PID)
    QString lockFileName = dataDir % QSL("/spoofer.lock");
    qDebug() << "lockFile:" << QDir::toNativeSeparators(lockFileName);
    lockfile = new QLockFile(lockFileName);
    lockfile->setStaleLockTime(0);
    bool retried = false; // we want to retry only once
retry:
    if (!lockfile->tryLock(0)) {
	qint64 pid;
	QString hostname, appname;
	sperr << "Error: \"" << lockFileName;
	switch (lockfile->error()) {
	    case QLockFile::LockFailedError:
		lockfile->getLockInfo(&pid, &hostname, &appname);
		sperr << "\" is locked by pid " << pid << " " << appname << endl;
#ifdef Q_OS_UNIX
		// On OSX, under conditions I haven't been able to reliably
		// reproduce, tryLock() may fail to remove a stale lock file.
		// (It may be related to unclean shutdown of launchd daemon.)
		if (!retried && kill(pid, 0) < 0 && errno == ESRCH) {
		    sperr << "But pid " << pid << " is not running." << endl;
		    if (lockfile->removeStaleLockFile()) {
			sperr << "Stale lock file removed." << endl;
			retried = true;
			goto retry;
		    } else {
			sperr << "Failed to remove stale lock file: " <<
			    strerror(errno) << endl;
		    }
		}
#endif
		break;
	    case QLockFile::PermissionError:
		sperr << "\": permission error" << endl;
		break;
	    default:
		sperr << "\": unknown error" <<
#ifdef Q_OS_UNIX
		    " (" << strerror(errno) << ")" <<
#endif
		    endl;
		break;
	}
	return false;
    }

    if (!initSignals()) return false;

    {
	// load history
	config->settings->beginGroup(QSL("history"));
	QStringList groups = config->settings->childGroups();
	foreach (QString group, groups) {
	    QString subnetstr = group;
	    subnetstr.replace(QSL(";"),QSL("/")); // QSettings doesn't allow "/" in keys
	    if (!subnetstr.contains(QSL("/"))) {
		qDebug() << "removing non-subnet group:" << group;
		config->settings->remove(group);
		continue;
	    }
	    SubnetAddr subnet(QHostAddress::parseSubnet(subnetstr));
	    if (subnet.pfxlen() <= 0) {
		qDebug() << "skipping invalid subnet group:" << group;
		continue;
	    }
	    config->settings->beginGroup(group);
	    RunRecord &rr = pastRuns[subnet];
	    rr.t = config->settings->value(QSL("time")).toLongLong();
	    rr.errors = config->settings->value(QSL("errors")).toUInt();
	    config->settings->endGroup();
	}
	config->settings->endGroup();
    }

    if (!(privServer = listen(true)))
	return false;
#ifndef EVERYONE_IS_PRIVILEGED
    if (!(unprivServer = listen(false)))
	return false;
#endif

    config->dataDir(dataDir);
    config->schedulerSocketName(privServer->fullServerName());

    config->sync();

    ncm = new QNetworkConfigurationManager(this);

    connect(ncm, &QNetworkConfigurationManager::configurationChanged,
	this, &App::handleNetChange);
    connect(ncm, &QNetworkConfigurationManager::configurationAdded,
	this, &App::handleNetChange);
    connect(ncm, &QNetworkConfigurationManager::configurationRemoved,
	this, &App::handleNetChange);

#if 0 // DEBUGGING: simulate frequent network change signal from OS or Qt
    QTimer *thrashTimer = new QTimer();
    connect(thrashTimer, &QTimer::timeout, this, &App::handleNetChange);
    thrashTimer->start(1000*10);
#endif

    proberTimer.setSingleShot(true);
    connect(&proberTimer, &QTimer::timeout, this, &App::startProber);

    netPollTimer.setSingleShot(true);
    connect(&netPollTimer, &QTimer::timeout, this, &App::scheduleNextProber);

    if (!paused) {
	// Wait 1s (in case of a burst of changes), then scheduleNextProber()
	netPollTimer.start(1000*1);
    }

    return true; // caller should continue with app.exec()
}

void App::cleanup()
{
    if (prober) {
	delete prober;
	prober = nullptr;
    }
    if (config)
	config->schedulerSocketName.remove();
    if (unprivServer) {
	delete unprivServer;
	unprivServer = nullptr;
    }
    if (privServer) {
	delete privServer;
	privServer = nullptr;
    }
    if (lockfile) {
	delete lockfile;
	lockfile = nullptr;
    }
}

App::~App()
{
    cleanup();
}

void App::uiAccept()
{
    QLocalServer *server = static_cast<QLocalServer*>(QObject::sender());
    bool privileged = (server == privServer);
    const char *label = privileged ? "privileged" : "unprivileged";
    qDebug() << "Scheduler: new" << label << "connection detected on" <<
	server->fullServerName();
    QLocalSocket *ui = server->nextPendingConnection();
    if (!ui) {
	qDebug() << "SCHEDULER ERROR: connection:" << server->errorString();
	return;
    }
    qDebug() << "Scheduler: accepted" << label << "connection";
    connect(ui, &QLocalSocket::readyRead, this, &App::uiRead);
    connect(ui, &QLocalSocket::disconnected, this, &App::uiDelete);
    uiSet.insert(ui);

    if (opAllowed(ui, config->unprivView)) {
	if (paused)
	    QDataStream(ui) << (qint32)SC_PAUSED;

	if (prober) {
	    // notify new UI of prober in progress
	    const sc_msg_text msg = { proberOutputFileName };
	    QDataStream(ui) << (qint32)SC_PROBER_STARTED << msg;
	} else if (nextProberStart.when) {
	    // notify new UI of next scheduled prober
	    QDataStream(ui) << (qint32)SC_SCHEDULED << nextProberStart;
	}
    }
}

void App::uiDelete()
{
    QLocalSocket *ui = static_cast<QLocalSocket*>(sender());
    qDebug() << "Scheduler: UI disconnected";
    uiSet.remove(ui);
    ui->deleteLater();
}

bool App::opAllowed(QLocalSocket *ui, const Config::MemberBase &cfgItem)
{
    if (ui->parent() == privServer) return true;
    if (cfgItem.variant().toBool()) return true;
    return false;
}

bool App::opAllowedVerbose(QLocalSocket *ui, const Config::MemberBase &cfgItem)
{
    if (opAllowed(ui, cfgItem)) return true;
    qDebug() << "Permission denied.";
    const sc_msg_text msg = { QSL("Permission denied.") };
    QDataStream(ui) << (qint32)SC_ERROR << msg;
    return false;
}

void App::uiRead()
{
    static QRegularExpression set_re(QSL("set\\s+(\\w+)\\s+(.*)"));
    QRegularExpressionMatch match;

    qDebug() << "Scheduler: uiRead()";

    QLocalSocket *ui = static_cast<QLocalSocket*>(sender());

    char data[1024];
    qint64 n;
    while ((n = ui->readLine(data, sizeof(data) - 1)) > 0) {
	while (n > 0 && data[n-1] == '\n') data[--n] = '\0';
	qDebug() << "Scheduler: UI read:" << data;
	if (strcmp(data, "shutdown") == 0) {
	    qDebug() << "Scheduler: shutdown by command";
	    this->shutdown();
	} else if (strcmp(data, "run") == 0) {
	    if (!opAllowedVerbose(ui, config->unprivTest)) continue;
	    if (prober) {
		const sc_msg_text msg =
		    { QSL("There is already a prober running.") };
		QDataStream(ui) << (qint32)SC_ERROR << msg;
	    } else {
		startProber();
	    }
	} else if (strcmp(data, "abort") == 0) {
	    if (!opAllowedVerbose(ui, config->unprivTest)) continue;
	    if (!prober) {
		const sc_msg_text msg = { QSL("There is no prober running.") };
		QDataStream(ui) << (qint32)SC_ERROR << msg;
	    } else {
		killProber();
		// const sc_msg_text msg = { QSL("Killing prober.") };
		// QDataStream(ui) << (qint32)SC_TEXT << msg;
	    }
	} else if (strcmp(data, "pause") == 0) {
	    if (!opAllowedVerbose(ui, config->unprivPref)) continue;
	    pause();
	    QDataStream(ui) << (qint32)SC_DONE_CMD;
	} else if (strcmp(data, "resume") == 0) {
	    if (!opAllowedVerbose(ui, config->unprivPref)) continue;
	    resume();
	    QDataStream(ui) << (qint32)SC_DONE_CMD;
	} else if ((match = set_re.match(QString::fromLocal8Bit(data))).hasMatch()) {
	    if (!opAllowedVerbose(ui, config->unprivPref)) continue;
	    QString name = match.captured(1);
	    QString value = match.captured(2);
	    qDebug() << "set_re: name =" << name << ", value =" << value;
	    SpooferBase::Config::MemberBase *cfgItem = config->find(name);
	    if (cfgItem) {
		if (!cfgItem->setFromString(value)) {
		    const sc_msg_text msg =
			{ QSL("\"%1\" is not a legal value for %2.")
			.arg(value, name) };
		    QDataStream(ui) << (qint32)SC_ERROR << msg;
		} else {
		    config->sync();
		    QDataStream(ui) << (qint32)SC_DONE_CMD;
		    foreach (QLocalSocket *ui, uiSet) {
			QDataStream(ui) << (qint32)SC_CONFIG_CHANGED;
		    }
		    // scheduleNextProber() in case any scheduling parameters
		    // changed (but wait 1s in case of a burst of changes)
		    scheduledAddrs.clear(); // force rescheduling
		    netPollTimer.start(1000*1);
		}
	    } else {
		const sc_msg_text msg =
		    { QSL("No such setting \"%1\".").arg(name) };
		QDataStream(ui) << (qint32)SC_ERROR << msg;
	    }
	} else if (strcmp(data, "sync") == 0) {
	    if (!opAllowedVerbose(ui, config->unprivPref)) continue;
	    config->sync();
	    QDataStream(ui) << (qint32)SC_DONE_CMD;
	} else if (strcmp(data, "dumpnetcfg") == 0) {
	    dumpNetCfg();
	    QDataStream(ui) << (qint32)SC_DONE_CMD;
	} else {
	    qDebug() << "Unknown command:" << data;
	    const sc_msg_text msg = { QSL("Unknown command.") };
	    QDataStream(ui) << (qint32)SC_ERROR << msg;
	}
    }
    if (n < 0) {
	qDebug() << "Scheduler: UI read error:" << ui->errorString();
    }
}

void App::startProber()
{
    proberTimer.stop();
    nextProberStart.when = 0;
    if (prober) {
	qDebug() << "startProber: there is already a prober running.";
	return;
    }
    QDir dir(dataDir);
    proberOutputFileName = dir.filePath(ftime_utc(proberLogFtime));
    sperr << "startProber: " << QDir::toNativeSeparators(proberOutputFileName)
	<< endl;

    prober = new QProcess(this);
    prober->setProcessChannelMode(QProcess::MergedChannels); // ... 2>&1
    prober->setStandardInputFile(QProcess::nullDevice()); // ... </dev/null
    prober->setStandardOutputFile(proberOutputFileName,
	QIODevice::Text | QIODevice::Truncate | QIODevice::WriteOnly); // ... >file
    QStringList args;
#if DEBUG
    if (config->useDevServer()) args << QSL("-T");
    if (config->pretendMode()) args << QSL("-p");
    if (config->standaloneMode()) args << QSL("-S");
#endif
    args << (config->shareData() ? QSL("-s1") : QSL("-s0"));
    if (config->enableIPv4()) args << QSL("-4");
    if (config->enableIPv6()) args << QSL("-6");
    if (optStandalone) args << QSL("-S");

    void (QProcess::*(qprocess_error))(QProcess::ProcessError) =
	&QProcess::error; // disambiguate overloaded function
    void (QProcess::*(qprocess_finished))(int,QProcess::ExitStatus)=
	&QProcess::finished; // disambiguate overloaded function
    connect(prober, &QProcess::started, this, &App::proberStarted);
    connect(prober, qprocess_error, this, &App::proberError);
    connect(prober, qprocess_finished, this, &App::proberFinished);

    QString proberPath = appDir % QSL("/spoofer-prober");
    prober->start(proberPath, args);
    // Unfortunately, if this process exits, ~QProcess() will kill the prober
    // process.  QProcess::startDetached() would avoid that, but then we can't
    // set up its input/output and signals/slots.
}

void App::killProber()
{
    prober->terminate();
    if (!prober->waitForFinished(1000) &&
	prober->state() != QProcess::NotRunning)
    {
	qDebug() << "Scheduler: prober did not terminate; killing by force.";
	prober->kill();
    }
}

void App::proberStarted()
{
    spout << "Scheduler: prober started, pid " << prober->processId() << endl;
    const sc_msg_text msg = { proberOutputFileName };
    foreach (QLocalSocket *ui, uiSet) {
	if (opAllowed(ui, config->unprivView))
	    QDataStream(ui) << (qint32)SC_PROBER_STARTED << msg;
    }
}

void App::proberError(QProcess::ProcessError e)
{
    QString msg;
    switch (e) {
	case QProcess::FailedToStart: msg = QSL("failed to start ");  break;
	case QProcess::Crashed:       msg = QSL("crashed in ");       break;
	case QProcess::Timedout:      msg = QSL("timed out in ");     break;
	case QProcess::WriteError:    msg = QSL("write error in ");   break;
	case QProcess::ReadError:     msg = QSL("read error in ");    break;
	case QProcess::UnknownError:  msg = QSL("unknown error in "); break;
	default:                      msg = QSL("error in ");         break;
    }
    msg.append(prober->program() % QSL(": ") % prober->errorString());
    sperr << "Scheduler: prober error: " << msg << endl;
    if (prober->state() == QProcess::NotRunning) {
	sc_msg_text scmsg = { msg };
	foreach (QLocalSocket *ui, uiSet) {
	    if (opAllowed(ui, config->unprivView))
		QDataStream(ui) << (qint32)SC_PROBER_ERROR << scmsg;
	}
	qDebug() << "Scheduler: delete prober";
	prober->deleteLater();
	prober = nullptr;
	recordRun(false);
	scheduleNextProber();
    }
}

void App::proberFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    char buf[80];
    bool success = false;
    qDebug() << "Scheduler: prober finished" << qPrintable(ftime_utc());

    sc_msg_text msg = { QString() };
    if (exitStatus == QProcess::NormalExit) {
	qDebug() << "Scheduler: prober exited normally, exit code " << exitCode;
	sprintf(buf, "normally (exit code %d).", exitCode);
	msg.text = QString::fromLocal8Bit(buf);
	success = (exitCode == 0);
    } else {
	qDebug() << "Scheduler: prober exited abnormally";
	msg.text = QSL("abnormally.");
    }
    foreach (QLocalSocket *ui, uiSet) {
	if (opAllowed(ui, config->unprivView))
	    QDataStream(ui) << (qint32)SC_PROBER_FINISHED << msg;
    }
    delete prober;
    prober = nullptr;

    recordRun(success);
    scheduleNextProber();

    QDir dir(dataDir);
    int nlogs = config->keepLogs();
    if (nlogs > 0) {
	QStringList lognames = dir.entryList(QStringList() << proberLogGlob,
	    QDir::Files, QDir::Name);
	for (int i = 0; i < lognames.size() - nlogs; i++) {
	    qDebug() << "removing old log" << lognames.at(i);
	    dir.remove(lognames.at(i));
	}
    }
}

void App::recordRun(bool success)
{
    QList<SubnetAddr> addrs = getAddresses();
    time_t now;
    time(&now);
    config->settings->beginGroup(QSL("history"));
    foreach (SubnetAddr addr, addrs) {
	SubnetAddr subnet = addr.prefix();
	RunRecord &rr = pastRuns[subnet];
	if (rr.t == now) continue; // don't duplicate
	rr.t = now;
	if (success)
	    rr.errors = 0;
	else
	    rr.errors++;
	config->settings->beginGroup(subnet.toString().replace(QSL("/"),QSL(";"))); // QSettings doesn't allow "/" in keys
	config->settings->setValue(QSL("time"), (qlonglong)rr.t);
	config->settings->setValue(QSL("errors"), rr.errors);
	config->settings->endGroup();
    }
    config->settings->endGroup();
    config->sync();
}

void App::pause()
{
    if (paused) return;
    qDebug() << "Scheduler: pausing";
    paused = true;
    proberTimer.stop();
    netPollTimer.stop();
    nextProberStart.when = 0;
    scheduledAddrs.clear();
    foreach (QLocalSocket *ui, uiSet) {
	if (opAllowed(ui, config->unprivView))
	    QDataStream(ui) << (qint32)SC_PAUSED;
    }
}

void App::resume()
{
    if (!paused) return;
    qDebug() << "Scheduler: resuming";
    paused = false;
    foreach (QLocalSocket *ui, uiSet) {
	if (opAllowed(ui, config->unprivView))
	    QDataStream(ui) << (qint32)SC_RESUMED;
    }
    scheduleNextProber();
}
